;; Generated package description from hyperbole.el
(define-package "hyperbole" "6.0.1" "GNU Hyperbole: The Everyday Hypertextual Information Manager" '((emacs "24.4")) :url "http://www.gnu.org/software/hyperbole" :keywords '("comm" "convenience" "files" "frames" "hypermedia" "languages" "mail" "matching" "mouse" "multimedia" "outlines" "tools" "wp"))
